/*
* BinarySearchTree
* 
* Implements a binary search tree, with AVL-specific functions (the main functions do NOT 
* try to maintain any kind of equilibrium). 
*
* Note: the sorting function (functions subQS*) were written for the first project and have been 
* copied here, with only minor modifications and simplifications due to the less versatile nature
* of the use. 
*
* Thibaut Cuvelier <tcuvelier@student.ulg.ac.be>
*
* 
* Philosophical definition:
*   program (pro'-gram) [n]
*   A magic spell cast over a computer allowing it to turn one's input into error messages. 
*/

#include "BinarySearchTree.h"   // Interface to implement
#include <stdlib.h>             // malloc() & free()
#include <stdio.h>              // fprintf()

/***
** Prototypes 
**/

typedef struct BinarySearchTreeNode_t Node;
typedef struct LinkedList_t List; 
typedef struct LinkedListNode_t ListNode; 

// Subprocedures related to trees. 
void   subFreeBinarySearchTree(Node*); 
Node*  subFindElement(Node*, int); 
Node*  subNewNode(int, const void*); 
Node*  subFindMaximum(Node* node); 
Node*  subDeleteMaximum(Node*); 
Node*  subRemoveElement(Node*, int, const void**); 
Node*  subCreateBinarySearchTree(Node**, int*, const void**, size_t, size_t); 
size_t subGetRankOf(Node*, int); 
void   subGetElementsInRange(List*, Node*, int, int); 

// Auxiliary functions (nothing to do with trees).
void  subSort(int*, const void**, size_t); 
void  subQSQuickSort(int*, const void**, int, int);
int   subQSPartition(int*, const void**, int, int);
void  subQSSwap(int*, const void**, int, int); 
List* subListNew(void); 
void  subListAdd(List*, const void*); 
void  subListPopHead(List*, const void**); 
void  subListFree(List*); 

/***
** Structures (opaque or hidden). 
**/

/** 
* Base opaque structure; is not part of the interface. 
*
* Contains a pointer to the root element and the size (in order to allow the getNumElements 
* function to have a constant time complexity, whilst adding only constants to other functions' 
* complexity). 
*/ 
struct BinarySearchTree_t {
    Node*  root;  /**< the root */
    size_t size;  /**< the size, equivalent to root->nNodes */
}; 

/** 
* Node structure; is not part of the interface, its existence should not be known outside this 
* file. 
* 
* Contains a pointer to the parent, and the left and right children; also keeps data (a key and 
* the proper data). 
*/
struct BinarySearchTreeNode_t {
    /** 
    * The data structure. 
    */
    /*@{*/
    Node*  parent;      /**< the parent */
    Node*  left;        /**< the left child */
    Node*  right;       /**< the right child */
    size_t nNodes;      /**< the number of nodes in this subtree */
    /*@}*/

    /** 
    * The data. 
    */
    /*@{*/
    int         key;    /**< the key */
    const void* data;   /**< the proper data */
    /*@}*/
}; 

/** Linked list; used for getElementsInRange. */
struct LinkedList_t {
    size_t size;    /**< the size */
    ListNode* head; /**< the head */
    ListNode* tail; /**< the tail */
}; 

/** Linked list node; used for getElementsInRange. */
struct LinkedListNode_t {
    const void* value; /**< the stored data */
    ListNode* next;    /**< the next value in the list */
    ListNode* prev;    /**< the previous value in the list */
};



/***
** Implementation
**/

BinarySearchTree* createEmptyBinarySearchTree(void) {
    BinarySearchTree* tree = (BinarySearchTree*) malloc(sizeof(BinarySearchTree)); 
    // malloc may return NULL, in case of problem. This value will do as return in case of error. 

    if(tree) {
        tree->root  = NULL; 
        tree->size  = 0; 
    }

    return tree; 
}

void freeBinarySearchTree(BinarySearchTree* tree) {
    subFreeBinarySearchTree(tree->root); 
    free(tree); 
}

size_t getNumElements(const BinarySearchTree* tree) {
    return tree->size; 
}

const void* findElement(const BinarySearchTree* tree, int key) {
    // If the tree is empty, the element is not inside. The condition will not fail: by lazy 
    // execution, one will try to access root only if tree is not NULL. 
    if(! tree || ! tree->root) {
        fprintf(stderr, 
            "[findElement][tree:%p][key:%d]: Asked to find an element in an mpty tree.", 
            (void*) tree, key); 
        return NULL; 
    }

    // Otherwise, start the recursive search (implemented in subFindElement). 
    Node* found = subFindElement(tree->root, key); 
    if(found != NULL) {
        return found->data; 
    } 

    return NULL; 
}

const void* removeElement(BinarySearchTree* tree, int key) {
    // Really similar to findElement; "just" deletes it once it's found. 
    if(tree == NULL || tree->root == NULL) {
        fprintf(stderr, 
            "[removeElement][tree:%p][key:%d]: Asked to remove an element from an empty tree.", 
            (void*) tree, key); 
        return NULL; 
    }

    const void* result = NULL; 
    tree->root = subRemoveElement(tree->root, key, &result); 
    return result; 
}

const void* insertElement(BinarySearchTree* tree, int key, const void* value) {
    // If there's no tree, stop and return NULL. 
    if(! tree) {
        fprintf(stderr, 
            "[insertElement][key:%d][value:%p]: Asked to insert an element in a non-existant "
            "tree.", key, value); 
        return NULL; 
    }

    Node* insert = subNewNode(key, value); 

    // Something went wrong with malloc(). Cannot go on. 
    if(! insert) {
        fprintf(stderr, 
            "[insertElement][tree:%p][key:%d][value:%p]: Asked to insert an element in an "
            "existant tree, but allocation failed.", (void*) tree, key, value); 
        return NULL; 
    }

    // If the tree is empty, it's quite easy, nothing to replace. If malloc fails, the tree will 
    // remain untouched - not true for the rest of the function. 
    // If the test above was put below, the tree would have remained in the same state, a bunch
    // of useless operations would have taken place. 
    if(! tree->root) {
        tree->root = insert; 
        tree->size = 1; 
        insert->nNodes = 0; 
        return NULL; 
    }

    // Otherwise, everything seems alright, go down the tree, find the place where the element 
    // should be inserted. 
    Node* first  = tree->root;  // Eventually, the place to insert the new element. 
    Node* second = NULL;        // First's parent. NULL == first->parent. 
    const void* result = NULL;  // The element to return. 

    while(first) {
        second = first; 
        if(insert->key <= first->key) {
            first = first->left; 
        } else {
            first = first->right; 
        }
    }

    if(! second) {
        // The tree is not empty, but the root may be replaced! 
        Node* toDelete = tree->root; 
        result = tree->root->data; 
        insert->left = tree->root->left; 
        insert->right = tree->root->right; 
        tree->root = insert; 
        free(toDelete); 
    } else if(insert->key == second->key) {
        // Update the old node by the new data (no one else is supposed to have access to 
        // internals, thanks to opaque structure; if this breaks something, that thing was not
        // supposed to exist!). 
        result = second->data; 
        second->key = key; 
        second->data = value; 
        free(insert); 
    } else {
        if(first) {
            result = first->data; 
            free(first); 
        } else {
            tree->size += 1; 
        }

        insert->parent = second;

        if(insert->key < second->key) {
            second->left = insert; 
        } else {
            second->right = insert; 
        }
    }

    second->nNodes = 1; 
    if(second->left) {
        second->nNodes += second->left->nNodes;
    }
    if(second->right) {
        second->nNodes += second->right->nNodes;
    }

    return result; 
}

BinarySearchTree* createBinarySearchTree(int* keys, const void** values, size_t numElements) {
    BinarySearchTree* tree = (BinarySearchTree*) malloc(sizeof(BinarySearchTree)); 

    if(! tree) {
        return NULL; 
    }

    // Sort the two array based on keys (if two elements are swapped in keys, the corresponding 
    // elements must be swapped in values). 
    subSort(keys, values, numElements); 
    // Note that inserting those keys now in the tree will cause it to be completely unbalanced!

    // Recurse: the center of the array becomes the root, the recursion on the left part will 
    // give left child (as the array is sorted, it is assured to be lower than root); on the right, 
    // the right child (same reason). Thus, the algorithm is correct. It is based on the divide-
    // and-conquer principle, as quick sort. 
    if(! subCreateBinarySearchTree(&(tree->root), keys, values, 0, numElements - 1)) {
        return NULL; 
    }

    tree->size  = numElements; 

    return tree; 
}

int getRankOf(const BinarySearchTree* tree, int key) {
    if(! tree || ! tree->root) {
        return -1; 
    }

    if(findElement(tree, key) == NULL) {
        return -1; 
    }

    // Returns the number of key lower than the given key, it does not check the existence of 
    // the elements! 
    return (int) subGetRankOf(tree->root, key); 
}

size_t getElementsInRange(const BinarySearchTree* tree, int min, int max, const void*** results) {
    if(min > max) {
        fprintf(stderr, "[getElementsInRange][tree:%p][min:%d][max:%d]: Min lower than max.", 
            (void*) tree, min, max); 
        *results = NULL; 
        return 0; 
    }

    if(! tree) {
        fprintf(stderr, "[getElementsInRange][tree:%p][min:%d][max:%d]: Got a non-existant tree.", 
            (void*) tree, min, max); 
        *results = NULL; 
        return 0; 
    }

    List* list = subListNew(); 

    if(! list) {
        fprintf(stderr, 
            "[getElementsInRange][tree:%p][min:%d][max:%d]: Unable to allocate an empty list.", 
            (void*) list, min, max); 
        *results = NULL; 
        return 0; 
    }

    // Working function, gets everything in the list. 
    subGetElementsInRange(list, tree->root, min, max); 

    if(list->size > 0) {
        // Put it all in the returned array. 
        const void** vals = (const void**) malloc(list->size * sizeof(const void *)); 

        if(! vals) {
            subListFree(list); 
            fprintf(stderr, 
                "[getElementsInRange][tree:%p][min:%d][max:%d]: Unable to allocate a new array "
                "of size %d.", (void*) list, min, max, list->size * sizeof(const void *)); 
            *results = NULL; 
            return 0; 
        }

        for(size_t i = 0; i < list->size; i += 1) {
            subListPopHead(list, &(vals[i])); 
        }

        *results = vals; 

        size_t ret = list->size; 
        subListFree(list); 
        return ret; 
    } else {
        subListFree(list); 
        *results = NULL; 
        return 0;
    }
}

/***
** Implementation details (subroutines)
**/

// Deletes recursively a whole tree from a node (postfix walk: first delete children, then delete 
// self). 
void subFreeBinarySearchTree(Node* node) {
    if(node) {
        if(node->left) {
            subFreeBinarySearchTree(node->left); 
        }

        if(node->right) {
            subFreeBinarySearchTree(node->right); 
        }

        free(node); 
    }
}

// Allocates a new node with the given data and no children nor parent. 
Node* subNewNode(int key, const void* data) {
    Node* node   = (Node*) malloc(sizeof(Node)); 

    if(node) {
        node->key    = key; 
        node->data   = data; 
        node->left   = NULL; 
        node->right  = NULL; 
        node->parent = NULL;
        node->nNodes = 1; 
    }

    return node; 
}

// Recursive function to search the search binary tree. 
Node* subFindElement(Node* node, int key) {
    if(! node) {
        return NULL; 
    }

    if(node->key == key) {
        return node; 
    }
    // Test based on binary search tree properties: elements with a key inferior to key are stored
    // in the left subtree. 
    else if(key < node->key) {
        return subFindElement(node->left, key); 
    } else {
        return subFindElement(node->right, key); 
    }
}

// Find the node's maximum and returns it. 
Node* subFindMaximum(Node* node) {
    if(! node) {
        return NULL; 
    }

    // Benefit from BST property: the maximum must be on the right side. 
    while(node->right) {
        node = node->right; 
    }

    return node; 
}

// Deletes the maximum node. 
Node* subDeleteMaximum(Node* x) {
    if(! x->right) {
        Node* result = x->left;
        //free(x); 
        return result; 
    }
    
    x->right = subDeleteMaximum(x->right); 
    
    x->nNodes = 1;
    if(x->left) {
        x->nNodes += x->left->nNodes; 
    }
    if(x->right) {
        x->nNodes += x->right->nNodes; 
    }

    return x; 
}

// Do the boring and recursive job of removing an element. 
Node* subRemoveElement(Node* x, int key, const void** data) {
    if(! x) {
        return NULL; 
    }

    // Find the element. 
    if(key < x->key) {
        x->left = subRemoveElement(x->left, key, data); 
    } else if(key > x->key) {
        x->right = subRemoveElement(x->right, key, data); 
    } else {
        // If there's only one child, the task is easy. 
        if(x->right == NULL) {
            Node* result = x->left;
            free(x); 
            return result; 
        }
        if(x->left == NULL) {
            Node* result = x->right;
            free(x); 
            return result; 
        }

        // Otherwise, replace it by its predecessor. 
        Node* t = x; 
        x = subFindMaximum(t->left); 
        x->left = subDeleteMaximum(t->left); 
        x->right = t->right; 
        free(t); 

        *data = x->data; 
    }
    
    x->nNodes = 1;
    if(x->left) {
        x->nNodes += x->left->nNodes; 
    }
    if(x->right) {
        x->nNodes += x->right->nNodes; 
    }

    return x; 
}

// Recursive part of the creation of an AVL-tree: put the center as the node, recurse (left 
// and right). 
// Error management: return NULL whenever something went wrong, pass it to parent instances. 
// If the returned value of the first call is NULL, then something went wrong; the instanciated
// nodes will be free'd. Other return values have no particular meaning. 
Node* subCreateBinarySearchTree(Node** node, int* keys, const void** val, size_t min, size_t max) {
    size_t delta = max - min; 

    size_t q = min + delta / 2; 
    Node* newNode = subNewNode(keys[q], val[q]); 

    if(! newNode) {
        return NULL; 
    }

    newNode->nNodes = q; 
    *node = newNode; 

    if(delta > 1) {
        if(! subCreateBinarySearchTree(&(newNode->left), keys, val, min, q-1)) {
            free(newNode); 
            return NULL; 
        }

        if(! subCreateBinarySearchTree(&(newNode->right), keys, val, q+1, max)) {
            free(newNode); 
            return NULL; 
        }
    } else if(min != max) {
        newNode->right = subNewNode(keys[min+1], val[min+1]); 
        newNode->right->nNodes = min + 1; 
    }

    return newNode; 
}

// Recursive working part of getRankOf; returns the number of keys lower than the given one in
// node's subtrees. Does not check their existence. 
size_t subGetRankOf(Node* node, int key) {
    if(! node) {
        return 0; 
    }

    if(key < node->key) {
        // If the key is inferior to the current root, return the rank of the key in the left 
        // subtree. 
        return subGetRankOf(node->left, key); 
    } else if(key > node->key) {
        // If the key is greater than the current root, return the number of keys at left, plus 
        // one (self) and the rank of the key in the right subtree. 
        size_t nodeLeftSize = 0; 
        if(node->left) {
            nodeLeftSize = node->left->nNodes; 
        }
        return 1 + nodeLeftSize + subGetRankOf(node->right, key); 
    } else {
        // If the key is the one at the root, return the number of keys in the left subtree. 
        if(node->left) {
            return node->left->nNodes; 
        } else {
            return 0; 
        }
    }
}

// Adds to the the elements with key between min and max which are node's children. 
void subGetElementsInRange(List* list, Node* node, int min, int max) {
    if(! node) {
        return; 
    }

    if(! list) {
        fprintf(stderr, 
            "[subGetElementsInRange][node:%p][min:%d][max:%d]: Was given an non-existant list.", 
            (void*) node, min, max); 
        return; 
    }

    if(min < node->key) {
        subGetElementsInRange(list, node->left, min, max); 
    }

    if(min <= node->key && max >= node->key) {
        subListAdd(list, node->data); 
    }

    if(max > node->key) {
        subGetElementsInRange(list, node->right, min, max); 
    }
}

/***
** Auxiliary algorithms: sorting and linked list management. 
**/

// A sorting algorithm (here, QuickSort is used, but any other would do). The algorithm must be a 
// little adapted: both arrays are to be sorted (modify swap so that it swaps elements in both of 
// them). 
void subSort(int* keys, const void** values, size_t size) {
    // subQS* are used for internal use of this quick sort algorithm. 
    subQSQuickSort(keys, values, 0, (int) size - 1); 
}

// Applies recursively the QuickSort algorithm on the two arrays. 
void subQSQuickSort(int* keys, const void** values, int begin, int end) {
    if (begin < end) {
        int q = subQSPartition(keys, values, begin, end);
        subQSQuickSort(keys, values, begin, q - 1);
        subQSQuickSort(keys, values, q + 1, end);
    }
}

// Partition the two arrays, comparisons being made on keys. 
int subQSPartition(int* keys, const void** values, int begin, int end) {
    int i = begin - 1; 
    int pivot = keys[end]; 

    for (int j = begin; j < end; j++) {
        if (keys[j] <= pivot) {
            i++;
            subQSSwap(keys, values, i, j);
        }
    }

    subQSSwap(keys, values, i + 1, end);
    return i + 1;
}

// Swaps the elements i and j in the two arrays. 
void subQSSwap(int* keys, const void** values, int i, int j) {
    const void* value = values[i];
    values[i] = values[j];
    values[j] = value;

    const int key = keys[i];
    keys[i] = keys[j];
    keys[j] = key;
}

// Allocates a new linked list; may return NULL. 
List* subListNew() {
    List* result = (List*) malloc(sizeof(List));

    if(result) {
        result->head = NULL; 
        result->tail = NULL; 
        result->size = 0; 
    }

    return result; 
}

// Adds a new node to the linked list, at the head. 
void subListAdd(List* list, const void* value) {
    if(! list) {
        fprintf(stderr, "[subListAdd]: Was given an non-existant list."); 
        return; 
    }

    ListNode* node = (ListNode*) malloc(sizeof(ListNode)); 

    if(! node) {
        fprintf(stderr, "[subListAdd][list:%p]: Unable to allocate a new node.", (void*) list); 
        return; 
    }

    list->size += 1; 
    node->value = value; 

    if(! list->head) {
        list->head = node; 
        list->tail = node; 
        node->prev = NULL; 
        node->next = NULL; 
    } else {
        list->tail->next = node; 
        node->prev = list->tail; 
        node->next = NULL; 
        list->tail = node; 
    }
}

// Puts in value as the first element of the list. 
void subListPopHead(List* list, const void** value) {
    if(! list) {
        fprintf(stderr, "[subListPopTail]: Was given an non-existant list."); 
        return; 
    }

    if(! list->head) {
        fprintf(stderr, "[subListPopTail][list:%p]: Was given an empty list.", (void*) list); 
        return; 
    }

    ListNode* node = list->head; 
    (*value) = node->value; 
    list->head = node->next; 
    free(node); 
}

// Frees the whole list and its elements (if there are any). 
void subListFree(List* list) {
    if(list) {
        if(list->head) {
            ListNode* node = list->head; 
            ListNode* mode = node; // next node
            do {
                node = mode; 
                mode = node->next; 
                free(node); 
            } while(node);
        }
        free(list); 
    }
}

