/* 
 * File:   Clustering.cpp
 * Author: Thibaut Cuvelier
 */

#include <vector>
#include <algorithm>
#include <limits>
#include <map>
#include "Clustering.hpp"
#include "ColourSet.hpp"

#include <iostream>

using namespace std; 

Clustering::Clustering() {}

vector<Colour> Clustering::clusteredColours(vector<Colour> values, unsigned levels, const vector<Colour> & startingPoints) {
  values_ = values; 
  computeOptimum(levels, startingPoints);
  replaceByOptimum(); 
  return values_;
}

void Clustering::computeOptimum(unsigned levels, const vector<Colour> & imposedPoints) {
  vector<ColourSet> current(levels), previous(levels); // The algorithm has converged when previous and current state are identical.

  // First state: Imposed points and random things to fill in. 
  for(vector<Colour>::size_type i = 0; i < imposedPoints.size(); ++i) 
    current[i].centroid(imposedPoints[i]);
  for(vector<Colour>::size_type i = imposedPoints.size(); i < levels; ++i) 
    current[i].centroid(Colour::randomColour()); 

  // Find the best clustering with this starting point.
  do {
    // Erase previous state. 
    previous = current; 
    current = vector<ColourSet>(levels); 

    // It makes no sense to reuse the previous state, as almost everything may change (except near convergence, maybe):
    // both the size of the list and the elements it contains. 

    // Imposed points. 
    for(vector<Colour>::size_type i = 0; i < imposedPoints.size(); ++i) 
      current[i].centroid(imposedPoints[i]);

    // New starting point: the result of previous iteration (new centroids: computed with the previous clustering).
    for(vector<Colour>::size_type i = imposedPoints.size(); i < levels; ++i) 
      current[i].centroid(previous[i].computeNewCentroid()); 

    // Assign points to sets: for each point...
    for(vector<Colour>::size_type i = 0; i < values_.size(); ++i) {
      double minDistance = numeric_limits<double>::infinity();
      vector<ColourSet>::size_type set; 

      // test all the sets...
      for(vector<ColourSet>::size_type j = 0; j < current.size(); ++j) {
        // If the square operation becomes a bottleneck, consider using a Manhattan distance, if the algorithm
        // still keeps its convergence properties. 
        double distance = values_[j].squaredEuclideanDistanceTo(current[j].centroid());

        if(distance < minDistance) {
          minDistance = distance; 
          set = j; 
        }
      }

      // and add it to the closest set. 
      current[set].addColour(values_[i]);
    }
  } while(! equal(current.begin(), current.end(), previous.begin())); // Iterate until nothing moves. 

  optimum_ = current; 
}

void Clustering::replaceByOptimum() {
  map<Colour, Colour> match;

  // First fill a map of the found colours.
  // Why a map? It is much more efficient than finding the matching colour for each pixel by iterating through
  // the whole thing, as the map guarantees logarithmic search. 
  // A hash table such as C++11's std::unordered_map may be better for performances, but 
  // (1) this is C++03 and (2) it requires a hash object, thus much more code. 
  for(vector<ColourSet>::iterator itCS = optimum_.begin(); itCS != optimum_.end(); ++itCS) 
    for(list<Colour>::const_iterator itC = itCS->colours().begin(); itC != itCS->colours().end(); ++itC) 
      match[*itC] = itCS->centroid(); 

  // Then replace them in the array.
  for(vector<Colour>::iterator it = values_.begin(); it != values_.end(); ++it) 
    *it = match[*it]; 
}
