/** 
 * HeapSort
 */

#include <stddef.h>
#include <stdio.h>

/* Prototypes */

void swap(const void**, size_t, size_t);
void HeapMaxHeapify(const void**, size_t, size_t, int (*)(const void*, const void*)); 
void HeapBuildMaxHeap(const void**, size_t, int (*)(const void*, const void*)); 

/* Functions */

/**
 * Swaps the two items. 
 * @param arr			the array where the variables will be swapped
 * @param firstIndex	the first index to swap
 * @param secondIndex	the second index to swap
 */ 
void swap(const void** arr, size_t firstIndex, size_t secondIndex)
{
  const void* tmp = arr[firstIndex];
  arr[firstIndex] = arr[secondIndex];
  arr[secondIndex] = tmp;
}

/**
 * Restore heap property for one element: it must be superior to its sons. 
 * 
 * @param arr		the array representing the heap
 * @param u			the element to consider
 * @param heapSize	the heap size to use
 * @param compare	the comparator to use
 */ 
void HeapMaxHeapify(const void** arr, size_t i, size_t heapSize, int (*compare)(const void*, const void*))
{
	// The two sons: l and r. The greatest element found: largest. 
	size_t l = 2*(i)+1;
	size_t r = 2*(i)+2;
	size_t largest; 

	// Determine the maximum element. 
	if (l < heapSize && compare(arr[l], arr[i]) > 0) {
		largest = l; 
	} else {
		largest = i; 
	}

	if (r < heapSize && compare(arr[r], arr[largest]) > 0) {
		largest = r; 
	}

	// If the root is not maximum, swap and recurse. 
	if (largest != i) {
		swap(arr, i, largest); 
		HeapMaxHeapify(arr, largest, heapSize, compare); 
	}
}

/**
 * Makes a heap from an array in place: the greatest element is at the root, 
 * and so much for its two sons. 
 * 
 * @param arr		the array to make heap
 * @param size		the size of the array to consider
 * @param compare	the comparator to use
 */ 
void HeapBuildMaxHeap(const void** arr, size_t size, int (*compare)(const void*, const void*))
{
	for(size_t i = (size+1) / 2 - 1; i > 0; i--) {
		HeapMaxHeapify(arr, i, size, compare);
	}
	HeapMaxHeapify(arr, 0, size, compare);
}

/**
 * Heap sorts an array. 
 * 
 * @param arr		the array to sort
 * @param size		the length of the array
 * @param compare	the comparator to use
 */ 
void Sort(const void** arr, size_t length, int (*compare)(const void*, const void*))
{
	HeapBuildMaxHeap(arr, length, compare); 

	/* The idea: at each iteration, take the maximum element, put it at the end, 
	   reduce the size of the heap and restore the heap property. */
	for(size_t i = 0; i < length; i++) {
		swap(arr, 0, length - i - 1); 
		HeapMaxHeapify(arr, 0, length - i - 1, compare); 
	}
}
