/** 
* QuickSort
*/ 

#include <stdio.h>

// Used to deactivate GCC's warnings about unused parameters - other 
// implementations of the functions might use them. 
#if __GNUC__ > 2
# define UNUSED __attribute__((unused))
#endif

/* Prototypes */

void swap(const void** array, int firstIndex, int secondIndex);
void quickSort(const void** array, int, int, 
	           int (*)(const void*, const void*), 
			   size_t(*)(const void**, int, int, const void *, 
						 int (*)(const void*, const void*)), 
			   const void* (*)(const void**, int, int, 
							   int (*)(const void*, const void*)));
size_t quickPartition(const void**, int, int, const void*, 
					  int (*)(const void*, const void*)); 
const void* quickPivot(const void**, int, int, 
					   int (*)(const void*, const void*));

/* Functions */

/**
 * Swaps the two items. 
 * @param arr			the array where the variables will be swapped
 * @param firstIndex	the first index to swap
 * @param secondIndex	the second index to swap
 */ 
void swap(const void** arr, int firstIndex, int secondIndex)
{
	const void* tmp = arr[firstIndex];
	arr[firstIndex] = arr[secondIndex];
	arr[secondIndex] = tmp;
}

/**
 * Pivot choice: returns the pivot to use in the quick sort. 
 * 
 * @param arr			the array to find the pivot in
 * @param beginning		the first index of the part to consider
 * @param end			the last index of the part to consider
 * @param compare		the comparator to use
 */ 
const void* quickPivot(const void** arr, int beginning UNUSED, int end, 
					   int (*compare)(const void*, const void*) UNUSED)
{
	// Arbitrarily, return the last element. Other implementations may be
	// faster. 
	return arr[end]; 
}

/**
 * Partition algorithm: it places elements inferior to the pivot in the left
 * part of the array, the ones that are superior to it in the right side. 
 * 
 * @param arr			the array to apply the algorithm to
 * @param beginning		the first index to partition
 * @param end			the last index to partition
 * @param pivot			the pivot to use
 * @param compare		the comparator to use
 */ 
size_t quickPartition(const void** arr, int beginning, int end, const void* pivot, 
					  int (*compare)(const void*, const void*))
{
	int i = beginning - 1;

	for (int j = beginning; j < end; j++) {
		if (compare(arr[j], pivot) <= 0) {
			i++;
			swap(arr, i, j);
		}
	}

	swap(arr, i + 1, end);
	return i + 1;
}

/**
 * Quick sort algorithm, made to be as generic as possible (functors are used
 * for pivot choice and partitionning) - after all, quick sort allows for many
 * specific choices. Their complexity will affect this algoritm's perofrmances. 
 * 
 * @param arr			the array to apply the algorithm to
 * @param beginning		the first index to sort
 * @param end			the last index to sort
 * @param pivot			the pivot to use
 * @param compare		the comparator to use
 * @param partition		the partitionning algorithm to use
 * @param pivot			the pivot algorithm to use
 */ 
void quickSort(const void** arr, int beginning, int end,
			   int (*compare)(const void*, const void*),
			   size_t(*partition)(const void**, int, int, const void *, 
								  int (*)(const void*, const void*)), 
			   const void* (*pivot)(const void**, int, int, 
									int (*)(const void*, const void*)))
{
	if (beginning < end) {
		int q = partition(arr, beginning, end, pivot(arr, beginning, end, compare), compare);
		quickSort(arr, beginning, q - 1, compare, partition, pivot);
		quickSort(arr, q + 1, end, compare, partition, pivot);
	}
}

/**
 * Bootstrap function to start the quick sort. 
 * 
 * @param arr			the array to sort
 * @param length		the size of the array
 * @param compare		the comparator to use
 */ 
void Sort(const void** arr, size_t length, 
	      int (*comparator)(const void*, const void*))
{
	quickSort(arr, 0, length - 1, comparator, quickPartition, quickPivot);
}